#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Curses interface class."""

import functools
import getpass
import sys

from glances.events_list import glances_events
from glances.globals import MACOS, WINDOWS, disable, enable, nativestr, u
from glances.logger import logger
from glances.outputs.glances_colors import GlancesColors
from glances.outputs.glances_unicode import unicode_message
from glances.processes import glances_processes, sort_processes_stats_list
from glances.timer import Timer

# Import curses library for "normal" operating system
try:
    import curses
    import curses.panel
    from curses.textpad import Textbox
except ImportError:
    logger.critical("Curses module not found. Glances cannot start in standalone mode.")
    if WINDOWS:
        logger.critical("For Windows you can try installing windows-curses with pip install.")
    sys.exit(1)


class _GlancesCurses:
    """This class manages the curses display (and key pressed).

    Note: It is a private class, use GlancesCursesClient or GlancesCursesBrowser.
    """

    _hotkeys = {
        '\n': {'handler': '_handle_enter'},
        '0': {'switch': 'disable_irix'},
        '1': {'switch': 'percpu'},
        '2': {'switch': 'disable_left_sidebar'},
        '3': {'switch': 'disable_quicklook'},
        '4': {'handler': '_handle_quicklook'},
        '5': {'handler': '_handle_top_menu'},
        '6': {'switch': 'meangpu'},
        '7': {'switch': 'disable_npu'},
        '8': {'switch': 'disable_mpp'},
        '/': {'switch': 'process_short_name'},
        'a': {'sort_key': 'auto'},
        'A': {'switch': 'disable_amps'},
        'b': {'switch': 'byte'},
        'B': {'handler': '_handle_diskio_iops'},
        'c': {'sort_key': 'cpu_percent'},
        'C': {'switch': 'disable_cloud'},
        'd': {'switch': 'disable_diskio'},
        'D': {'switch': 'disable_containers'},
        # 'e' > Enable/Disable process extended
        'E': {'handler': '_handle_erase_filter'},
        'f': {'handler': '_handle_fs_stats'},
        'F': {'switch': 'fs_free_space'},
        'g': {'switch': 'generate_graph'},
        'G': {'switch': 'disable_gpu'},
        'h': {'switch': 'help_tag'},
        'i': {'sort_key': 'io_counters'},
        'I': {'switch': 'disable_ip'},
        'j': {'switch': 'programs'},
        # 'k' > Kill selected process
        'K': {'switch': 'disable_connections'},
        'l': {'switch': 'disable_alert'},
        'L': {'handler': '_handle_diskio_latency'},
        'm': {'sort_key': 'memory_percent'},
        'M': {'switch': 'reset_minmax_tag'},
        'n': {'switch': 'disable_network'},
        'N': {'switch': 'disable_now'},
        'o': {'sort_key': 'cpu_num'},
        'p': {'sort_key': 'name'},
        'P': {'switch': 'disable_ports'},
        # 'q' or ESCAPE > Quit
        'Q': {'switch': 'enable_irq'},
        'r': {'switch': 'disable_smart'},
        'R': {'switch': 'disable_raid'},
        's': {'switch': 'disable_sensors'},
        'S': {'switch': 'sparkline'},
        't': {'sort_key': 'cpu_times'},
        'T': {'switch': 'network_sum'},
        'u': {'sort_key': 'username'},
        'U': {'switch': 'network_cumul'},
        'V': {'switch': 'disable_vms'},
        'w': {'handler': '_handle_clean_logs'},
        'W': {'switch': 'disable_wifi'},
        'x': {'handler': '_handle_clean_critical_logs'},
        'z': {'handler': '_handle_disable_process'},
        '+': {'handler': '_handle_increase_nice'},
        '-': {'handler': '_handle_decrease_nice'},
        # "<" (shift + left arrow) navigation through process sort
        # ">" (shift + right arrow) navigation through process sort
        # "<" (left arrow) scroll through process name
        # ">" (right arrow) scroll through process name
        # 'UP' > Up in the server list
        # 'DOWN' > Down in the server list
    }

    _sort_loop = sort_processes_stats_list

    # Define top menu
    _top = ['quicklook', 'cpu', 'percpu', 'npu', 'mpp', 'gpu', 'mem', 'memswap', 'load']
    _quicklook_max_width = 48

    # Define left sidebar
    # This variable is used in the make webui task in order to generate the
    # glances/outputs/static/js/uiconfig.json file for the web interface
    # This list can also be overwritten by the configuration file ([outputs] left_menu option)
    _left_sidebar = [
        'network',
        'ports',
        'wifi',
        'connections',
        'diskio',
        'fs',
        'irq',
        'folders',
        'raid',
        'smart',
        'sensors',
        'now',
    ]
    _left_sidebar_min_width = 23
    _left_sidebar_max_width = 34

    # Define right sidebar in a method because it depends of self.args.programs
    # See def _right_sidebar method

    def __init__(self, config=None, args=None):
        # Init
        self.config = config
        self.args = args

        # Init windows positions
        self.term_w = 80
        self.term_h = 24

        # Space between stats
        self.space_between_column = 3
        self.space_between_line = 2

        # Init the curses screen
        try:
            self.screen = curses.initscr()
            if not self.screen:
                logger.critical("Cannot init the curses library.\n")
                sys.exit(1)
            else:
                logger.debug(f"Curses library initialized with term: {curses.longname()}")
        except Exception as e:
            if args.export:
                logger.info("Cannot init the curses library, quiet mode on and export.")
                args.quiet = True
                return

            logger.critical(f"Cannot init the curses library ({e})")
            sys.exit(1)

        # Load configuration file
        self.load_config(config)

        # Init Curses cursor
        self._init_curses_cursor()

        # Init the colors
        self.colors_list = GlancesColors(args).get()

        # Init main window
        self.term_window = self.screen.subwin(0, 0)

        # Init edit filter tag
        self.edit_filter = False

        # Init nice increase/decrease tag
        self.increase_nice_process = False
        self.decrease_nice_process = False

        # Init kill process tag
        self.kill_process = False

        # Init the process min/max reset
        self.args.reset_minmax_tag = False

        # Init Glances cursor
        self.args.cursor_position = 0
        self.args.cursor_process_name_position = 0

        # For the moment cursor only available in standalone mode
        self.args.disable_cursor = not self.args.is_standalone

        # Catch key pressed with non blocking mode
        self.term_window.keypad(1)
        self.term_window.nodelay(1)
        self.pressedkey = -1

        # Is this the end ?
        self.is_end = False

        # History tag
        self._init_history()

    def load_config(self, config):
        """Load the outputs section of the configuration file."""
        if config is not None and config.has_section('outputs'):
            logger.debug('Read the outputs section in the configuration file')
            # Separator
            self.args.enable_separator = config.get_bool_value(
                'outputs', 'separator', default=self.args.enable_separator
            )
            # Set the left sidebar list
            self._left_sidebar = config.get_list_value('outputs', 'left_menu', default=self._left_sidebar)
            # Background color
            self.args.disable_bg = config.get_bool_value('outputs', 'disable_bg', default=self.args.disable_bg)

    def _right_sidebar(self):
        return [
            'vms',
            'containers',
            'processcount',
            'amps',
            'programlist' if self.args.programs else 'processlist',
            'alert',
        ]

    def _init_history(self):
        """Init the history option."""

        self.reset_history_tag = False

    def _init_curses_cursor(self):
        """Init cursors."""

        if hasattr(curses, 'noecho'):
            curses.noecho()
        if hasattr(curses, 'cbreak'):
            curses.cbreak()
        self.set_cursor(0)

    def set_cursor(self, value):
        """Configure the curse cursor appearance.

        0: invisible
        1: visible
        2: very visible
        """
        if hasattr(curses, 'curs_set'):
            try:
                curses.curs_set(value)
            except Exception:
                pass

    def get_key(self, window):
        return window.getch()

    def catch_actions_from_hotkey(self, hotkey):
        if self.pressedkey == ord(hotkey) and 'switch' in self._hotkeys[hotkey]:
            self._handle_switch(hotkey)
        elif self.pressedkey == ord(hotkey) and 'sort_key' in self._hotkeys[hotkey]:
            self._handle_sort_key(hotkey)
        if self.pressedkey == ord(hotkey) and 'handler' in self._hotkeys[hotkey]:
            action = getattr(self, self._hotkeys[hotkey]['handler'])
            action()

    def catch_other_actions_maybe_return_to_browser(self, return_to_browser):
        {
            self.pressedkey in {ord('e')} and not self.args.programs: self._handle_process_extended,
            self.pressedkey in {ord('k')} and not self.args.disable_cursor: self._handle_kill_process,
            self.pressedkey
            in {curses.KEY_LEFT if self.args.arrow_keys_sort else curses.KEY_SLEFT}: self._handle_sort_left,
            self.pressedkey
            in {curses.KEY_RIGHT if self.args.arrow_keys_sort else curses.KEY_SRIGHT}: self._handle_sort_right,
            self.pressedkey
            in {curses.KEY_SLEFT if self.args.arrow_keys_sort else curses.KEY_LEFT}: self._handle_process_name_left,
            self.pressedkey
            in {curses.KEY_SRIGHT if self.args.arrow_keys_sort else curses.KEY_RIGHT}: self._handle_process_name_right,
            self.pressedkey in {curses.KEY_UP, 65} and not self.args.disable_cursor: self._handle_cursor_up,
            self.pressedkey in {curses.KEY_DOWN, 66} and not self.args.disable_cursor: self._handle_cursor_down,
            self.pressedkey in {curses.KEY_F5, 18}: self._handle_refresh,
            self.pressedkey in {ord('\x1b'), ord('q')}: functools.partial(self._handle_quit, return_to_browser),
        }.get(True, lambda: None)()

    def __catch_key(self, return_to_browser=False):
        # Catch the pressed key
        self.pressedkey = self.get_key(self.term_window)
        if self.pressedkey == -1:
            return self.pressedkey

        # Actions (available in the global hotkey dict)...
        logger.debug(f"Keypressed (code: {self.pressedkey})")
        [self.catch_actions_from_hotkey(hotkey) for hotkey in self._hotkeys]

        # Other actions with key > 255 (ord will not work) and/or additional test...
        self.catch_other_actions_maybe_return_to_browser(return_to_browser)

        # Return the key code
        return self.pressedkey

    def _handle_switch(self, hotkey):
        option = '_'.join(self._hotkeys[hotkey]['switch'].split('_')[1:])
        if self._hotkeys[hotkey]['switch'].startswith('disable_'):
            if getattr(self.args, self._hotkeys[hotkey]['switch']):
                enable(self.args, option)
            else:
                disable(self.args, option)
        elif self._hotkeys[hotkey]['switch'].startswith('enable_'):
            if getattr(self.args, self._hotkeys[hotkey]['switch']):
                disable(self.args, option)
            else:
                enable(self.args, option)
        else:
            setattr(
                self.args,
                self._hotkeys[hotkey]['switch'],
                not getattr(self.args, self._hotkeys[hotkey]['switch']),
            )

    def _handle_sort_key(self, hotkey):
        glances_processes.set_sort_key(self._hotkeys[hotkey]['sort_key'], self._hotkeys[hotkey]['sort_key'] == 'auto')

    def _handle_enter(self):
        self.edit_filter = not self.edit_filter

    def _handle_quicklook(self):
        self.args.full_quicklook = not self.args.full_quicklook
        if self.args.full_quicklook:
            self.enable_fullquicklook()
        else:
            self.disable_fullquicklook()

    def _handle_top_menu(self):
        self.args.disable_top = not self.args.disable_top
        if self.args.disable_top:
            self.disable_top()
        else:
            self.enable_top()

    def _handle_process_extended(self):
        self.args.enable_process_extended = not self.args.enable_process_extended
        if not self.args.enable_process_extended:
            glances_processes.disable_extended()
        else:
            glances_processes.enable_extended()
        self.args.disable_cursor = self.args.enable_process_extended and self.args.is_standalone

    def _handle_erase_filter(self):
        glances_processes.process_filter = None

    def _handle_fs_stats(self):
        self.args.disable_fs = not self.args.disable_fs
        self.args.disable_folders = not self.args.disable_folders

    def _handle_increase_nice(self):
        self.increase_nice_process = not self.increase_nice_process

    def _handle_decrease_nice(self):
        self.decrease_nice_process = not self.decrease_nice_process

    def _handle_kill_process(self):
        self.kill_process = not self.kill_process

    def _handle_process_name_left(self):
        if self.args.cursor_process_name_position > 0:
            self.args.cursor_process_name_position -= 1

    def _handle_process_name_right(self):
        self.args.cursor_process_name_position += 1

    def _handle_clean_logs(self):
        glances_events.clean()

    def _handle_clean_critical_logs(self):
        glances_events.clean(critical=True)

    def _handle_disable_process(self):
        self.args.disable_process = not self.args.disable_process
        if self.args.disable_process:
            glances_processes.disable()
        else:
            glances_processes.enable()

    def _handle_diskio_iops(self):
        """Switch between bytes/s and IOPS for Disk IO."""
        self.args.diskio_iops = not self.args.diskio_iops
        if self.args.diskio_iops:
            self.args.diskio_latency = False

    def _handle_diskio_latency(self):
        """Switch between bytes/s and latency for Disk IO."""
        self.args.diskio_latency = not self.args.diskio_latency
        if self.args.diskio_latency:
            self.args.diskio_iops = False

    def _handle_sort_left(self):
        next_sort = (self.loop_position() - 1) % len(self._sort_loop)
        glances_processes.set_sort_key(self._sort_loop[next_sort], False)

    def _handle_sort_right(self):
        next_sort = (self.loop_position() + 1) % len(self._sort_loop)
        glances_processes.set_sort_key(self._sort_loop[next_sort], False)

    def _handle_cursor_up(self):
        if self.args.cursor_position > 0:
            self.args.cursor_position -= 1

    def _handle_cursor_down(self):
        if self.args.cursor_position < glances_processes.processes_count:
            self.args.cursor_position += 1

    def _handle_quit(self, return_to_browser):
        if return_to_browser:
            logger.info("Stop Glances client and return to the browser")
        else:
            logger.info(f"Stop Glances (keypressed: {self.pressedkey})")
            # End the curses window
            self.end()
            # Exit the program
            sys.exit(0)

    def _handle_refresh(self):
        glances_processes.reset_internal_cache()

    def loop_position(self):
        """Return the current sort in the loop"""
        for i, v in enumerate(self._sort_loop):
            if v == glances_processes.sort_key:
                return i
        return 0

    def disable_top(self):
        """Disable the top panel"""
        for p in self._top:
            setattr(self.args, 'disable_' + p, True)

    def enable_top(self):
        """Enable the top panel"""
        for p in self._top:
            setattr(self.args, 'disable_' + p, False)

    def disable_fullquicklook(self):
        """Disable the full quicklook mode"""
        for p in ['quicklook', 'cpu', 'npu', 'mpp', 'gpu', 'mem', 'memswap']:
            setattr(self.args, 'disable_' + p, False)

    def enable_fullquicklook(self):
        """Disable the full quicklook mode"""
        self.args.disable_quicklook = False
        for p in ['cpu', 'npu', 'mpp', 'gpu', 'mem', 'memswap']:
            setattr(self.args, 'disable_' + p, True)

    def end(self):
        """Shutdown the curses window."""
        if hasattr(curses, 'echo'):
            curses.echo()
        if hasattr(curses, 'nocbreak'):
            curses.nocbreak()
        try:
            curses.curs_set(1)
        except Exception:
            pass
        try:
            curses.endwin()
        except Exception:
            pass
        self.is_end = True

    def init_line_column(self):
        """Init the line and column position for the curses interface."""
        self.init_line()
        self.init_column()

    def init_line(self):
        """Init the line position for the curses interface."""
        self.line = 0
        self.next_line = 0

    def init_column(self):
        """Init the column position for the curses interface."""
        self.column = 0
        self.next_column = 0

    def new_line(self, separator=False):
        """New line in the curses interface."""
        self.line = self.next_line

    def new_column(self):
        """New column in the curses interface."""
        self.column = self.next_column

    def separator_line(self, color='SEPARATOR'):
        """Add a separator line in the curses interface."""
        if not self.args.enable_separator:
            return
        self.new_line()
        self.line -= 1
        line_width = self.term_window.getmaxyx()[1] - self.column
        if self.line >= 0 and self.line < self.term_window.getmaxyx()[0]:
            position = [self.line, self.column]
            line_color = self.colors_list[color]
            line_type = curses.ACS_HLINE if not self.args.disable_unicode else unicode_message('MEDIUM_LINE', self.args)
            self.term_window.hline(
                *position,
                line_type,
                line_width,
                line_color,
            )

    def __get_stat_display(self, stats, layer):
        """Return a dict of dict with all the stats display.
        # TODO: Drop extra parameter

        :param stats: Global stats dict
        :param layer: ~ cs_status
            "None": standalone or server mode
            "Connected": Client is connected to a Glances server
            "SNMP": Client is connected to a SNMP server
            "Disconnected": Client is disconnected from the server

        :returns: dict of dict
            * key: plugin name
            * value: dict returned by the get_stats_display Plugin method
        """
        ret = {}

        for p in stats.getPluginsList(enable=False):
            # Ignore Quicklook because it is compute later in __display_top
            if p == 'quicklook':
                continue

            # Compute the plugin max size for the left sidebar
            plugin_max_width = None
            if p in self._left_sidebar:
                plugin_max_width = min(
                    self._left_sidebar_max_width,
                    max(self._left_sidebar_min_width, self.term_window.getmaxyx()[1] - 105),
                )

            # Get the view
            ret[p] = stats.get_plugin(p).get_stats_display(args=self.args, max_width=plugin_max_width)

        return ret

    def display(self, stats, cs_status=None):
        """Display stats on the screen.

        :param stats: Stats database to display
        :param cs_status:
            "None": standalone or server mode
            "Connected": Client is connected to a Glances server
            "SNMP": Client is connected to a SNMP server
            "Disconnected": Client is disconnected from the server

        :return: True if the stats have been displayed else False if the help have been displayed
        """
        # Init the internal line/column for Glances Curses
        self.init_line_column()

        # Update the stats messages
        ###########################

        # Get all the plugins view
        self.args.cs_status = cs_status
        __stat_display = self.__get_stat_display(stats, layer=cs_status)

        # Display the stats on the curses interface
        ###########################################

        # Help screen (on top of the other stats)
        if self.args.help_tag:
            # Display the stats...
            self.display_plugin(stats.get_plugin('help').get_stats_display(args=self.args))
            # ... and exit
            return False

        # =======================================
        # Display first line (system+ip+uptime)
        # Optionally: Cloud is on the second line
        # =======================================
        self.__display_header(__stat_display)
        self.separator_line()

        # ====================================================================
        # Display second line (<SUMMARY>+CPU|PERCPU+<NPU>+<GPU>+LOAD+MEM+SWAP)
        # ====================================================================
        self.__display_top(__stat_display, stats)
        self.init_column()
        self.separator_line()

        # ===================================================================
        # Display left sidebar (NETWORK+PORTS+DISKIO+FS+SENSORS+Current time)
        # ===================================================================
        self.__display_left(__stat_display)

        # ====================================
        # Display right stats (process and co)
        # ====================================
        self.__display_right(__stat_display)

        # =====================
        # Others popup messages
        # =====================

        # Display edit filter popup
        # Only in standalone mode (cs_status is None)
        if self.edit_filter and cs_status is None:
            new_filter = self.display_popup(
                'Process filter pattern: \n\n'
                + 'Examples:\n'
                + '- .*python.*\n'
                + '- /usr/lib.*\n'
                + '- name:.*nautilus.*\n'
                + '- cmdline:.*glances.*\n'
                + '- username:nicolargo\n'
                + '- username:^root        ',
                popup_type='input',
                input_value=glances_processes.process_filter_input,
            )
            glances_processes.process_filter = new_filter
        elif self.edit_filter and cs_status is not None:
            self.display_popup('Process filter only available in standalone mode')
        self.edit_filter = False

        # Manage increase/decrease nice level of the selected process
        # Only in standalone mode (cs_status is None)
        if self.increase_nice_process and cs_status is None:
            self.nice_increase(stats.get_plugin('processlist').get_raw()[self.args.cursor_position])
        self.increase_nice_process = False
        if self.decrease_nice_process and cs_status is None:
            self.nice_decrease(stats.get_plugin('processlist').get_raw()[self.args.cursor_position])
        self.decrease_nice_process = False

        # Display kill process confirmation popup
        # Only in standalone mode (cs_status is None)
        if self.kill_process and cs_status is None:
            self.kill(stats.get_plugin('processlist').get_raw()[self.args.cursor_position])
        elif self.kill_process and cs_status is not None:
            self.display_popup('Kill process only available for local processes')
        self.kill_process = False

        # Display graph generation popup
        if self.args.generate_graph:
            if 'graph' in stats.getExportsList():
                self.display_popup(f'Generate graph in {self.args.export_graph_path}')
            else:
                logger.warning('Graph export module is disable. Run Glances with --export graph to enable it.')
                self.args.generate_graph = False

        return True

    def nice_increase(self, process):
        glances_processes.nice_increase(process['pid'])

    def nice_decrease(self, process):
        glances_processes.nice_decrease(process['pid'])

    def kill(self, process):
        """Kill a process, or a list of process if the process has a childrens field.

        :param process
        :return: None
        """
        logger.debug(f"Selected process to kill: {process}")

        if 'childrens' in process:
            pid_to_kill = process['childrens']
        else:
            pid_to_kill = [process['pid']]

        confirm = self.display_popup(
            'Kill process: {} (pid: {}) ?\n\nConfirm ([y]es/[n]o): '.format(
                process['name'],
                ', '.join(map(str, pid_to_kill)),
            ),
            popup_type='yesno',
        )

        if confirm.lower().startswith('y'):
            for pid in pid_to_kill:
                try:
                    ret_kill = glances_processes.kill(pid)
                except Exception as e:
                    logger.error(f'Can not kill process {pid} ({e})')
                else:
                    logger.info(f'Kill signal has been sent to process {pid} (return code: {ret_kill})')

    def __display_header(self, stat_display):
        """Display the firsts lines (header) in the Curses interface.

        system + ip + uptime
        (cloud)
        """
        # First line
        self.new_line()
        l_uptime = 1
        for i in ['system', 'ip', 'uptime']:
            if i in stat_display:
                l_uptime += self.get_stats_display_width(stat_display[i])

        display_system_optional = self.term_window.getmaxyx()[1] >= l_uptime

        # Calculate the initial `space_between_column` based on displayed system messages
        msgs = stat_display["system"]["msgdict"]
        visible_msgs = [msg for msg in msgs if display_system_optional or not msg["optional"]]
        ends_with_space = bool(visible_msgs) and visible_msgs[-1]["msg"].endswith(" ")
        self.space_between_column = 0 if ends_with_space else 1

        self.display_plugin(stat_display["system"], display_optional=display_system_optional)
        self.space_between_column = 3
        if 'ip' in stat_display:
            self.new_column()
            self.display_plugin(stat_display["ip"], display_optional=(self.term_window.getmaxyx()[1] >= 100))
        self.new_column()
        cloud_width = self.get_stats_display_width(stat_display.get("cloud", 0))
        self.display_plugin(stat_display["uptime"], add_space=-(cloud_width != 0))
        self.init_column()
        if cloud_width != 0:
            # Second line (optional)
            self.new_line()
            self.display_plugin(stat_display["cloud"])

    def _get_plugin_width(self, stat_display):
        """
        Compute the display width for each top plugin.

        The width is calculated using the plugin display data currently
        available in stat_display. Plugins that are not enabled through
        runtime arguments are assigned a width of 0.
        """
        # Store calculated widths for each plugin
        plugin_widths = {}

        for p in self._top:
            plugin_widths[p] = (
                self.get_stats_display_width(stat_display.get(p, 0)) if hasattr(self.args, 'disable_' + p) else 0
            )

        return plugin_widths

    def _get_stats_summary(self, stat_display, plugin_widths):
        """
        Compute total plugin width and number of active plugins.

        Returns
        -------
        tuple
                stats_width (int): Combined width of all plugins.
                stats_number (int): Number of plugins containing
                    displayable content.

        """
        # Compute total width occupied by all plugins
        stats_width = sum(plugin_widths.values())

        # Count plugins that currently contain display data
        # Quicklook is excluded from this stage of computation
        stats_number = sum(
            [int(stat_display[p]['msgdict'] != []) for p in self._top if not getattr(self.args, 'disable_' + p)]
        )

        return stats_width, stats_number

    def _handle_quicklook_for_display(
        self,
        stat_display,
        stats,
        plugin_widths,
        stats_width,
        stats_number,
    ):
        """
        Compute, retrieve, and display the quicklook plugin.

        This helper handles quicklook width calculation, plugin retrieval,
        display rendering, and spacing adjustments required before the
        remaining top plugins are rendered.
        """
        # Compute available width for quicklook display
        if self.args.full_quicklook:
            quicklook_width = self.term_window.getmaxyx()[1] - (
                stats_width + 8 + stats_number * self.space_between_column
            )
        else:
            quicklook_width = min(
                self.term_window.getmaxyx()[1] - (stats_width + 8 + stats_number * self.space_between_column),
                self._quicklook_max_width - 5,
            )

        try:
            # Retrieve quicklook display data
            stat_display["quicklook"] = stats.get_plugin('quicklook').get_stats_display(
                max_width=quicklook_width,
                args=self.args,
            )

        except AttributeError as e:
            logger.debug(f"Quicklook plugin not available ({e})")

        else:
            # Update quicklook width after successful retrieval
            plugin_widths['quicklook'] = self.get_stats_display_width(stat_display["quicklook"])

            # Recompute total width including quicklook
            stats_width = sum(plugin_widths.values()) + 1

        # Quicklook uses a dedicated single-column spacing rule
        self.space_between_column = 1

        # Display quicklook before remaining plugins
        self.display_plugin(stat_display["quicklook"])
        self.new_column()

        return plugin_widths, stats_width

    def _compute_spacing_and_optional(
        self,
        stat_display,
        plugin_widths,
        stats_width,
        stats_number,
    ):
        """
        Compute spacing and optional display behaviour for top plugins.

        This helper adjusts spacing dynamically based on terminal width and
        disables optional CPU and MEM display elements when horizontal
        space becomes constrained.
        """
        # Enable optional display for all plugins by default
        plugin_display_optional = dict.fromkeys(self._top, True)

        if stats_number > 1:
            # Compute spacing between plugin columns
            self.space_between_column = max(
                1,
                int((self.term_window.getmaxyx()[1] - stats_width) / (stats_number - 1)),
            )

            for p in ['mem', 'cpu']:
                # Remove optional display elements when spacing becomes too small
                if self.space_between_column < 3:
                    plugin_display_optional[p] = False

                    # Recompute plugin width without optional display elements
                    plugin_widths[p] = (
                        self.get_stats_display_width(
                            stat_display[p],
                            without_option=True,
                        )
                        if hasattr(self.args, 'disable_' + p)
                        else 0
                    )

                    # Recompute total width after optional removal
                    stats_width = sum(plugin_widths.values()) + 1

                    # Recompute spacing with updated widths
                    self.space_between_column = max(
                        1,
                        int((self.term_window.getmaxyx()[1] - stats_width) / (stats_number - 1)),
                    )
        else:
            # No spacing required when only one plugin is displayed
            self.space_between_column = 0

        return plugin_display_optional, plugin_widths

    def __display_top(self, stat_display, stats):
        """Display the second line in the Curses interface.

        <QUICKLOOK> + CPU|PERCPU + <NPU> + <GPU> + MEM + SWAP + LOAD
        """
        self.init_column()
        self.new_line()

        # Init quicklook
        stat_display['quicklook'] = {'msgdict': []}

        # Get Dict for plugins width
        plugin_widths = self._get_plugin_width(stat_display)

        # Get width of all plugins and number of plugin but quicklook
        stats_width, stats_number = self._get_stats_summary(stat_display, plugin_widths)

        if not self.args.disable_quicklook:
            plugin_widths, stats_width = self._handle_quicklook_for_display(
                stat_display, stats, plugin_widths, stats_width, stats_number
            )

        # Compute spaces between plugins
        # Note: Only one space between Quicklook and others

        plugin_display_optional, plugin_widths = self._compute_spacing_and_optional(
            stat_display, plugin_widths, stats_width, stats_number
        )

        # Display CPU, MEM, SWAP and LOAD
        for p in self._top:
            if p == 'quicklook':
                continue
            if p in stat_display:
                self.display_plugin(stat_display[p], display_optional=plugin_display_optional[p])
            if p != 'load':
                # Skip last column
                self.new_column()

        # Space between column
        self.space_between_column = 3

        # Backup line position
        self.saved_line = self.next_line

    def __display_left(self, stat_display):
        """Display the left sidebar in the Curses interface."""
        self.init_column()

        if self.args.disable_left_sidebar:
            return

        for p in self._left_sidebar:
            if (hasattr(self.args, 'enable_' + p) or hasattr(self.args, 'disable_' + p)) and p in stat_display:
                self.new_line()
                if p == 'sensors':
                    self.display_plugin(
                        stat_display['sensors'],
                        max_y=(self.term_window.getmaxyx()[0] - self.get_stats_display_height(stat_display['now']) - 2),
                    )
                else:
                    self.display_plugin(stat_display[p])

    def __display_right(self, stat_display):
        """Display the right sidebar in the Curses interface.

        docker + processcount + amps + processlist + alert
        """
        # Do not display anything if space is not available...
        if self.term_window.getmaxyx()[1] < self._left_sidebar_min_width:
            return

        # Restore line position
        self.next_line = self.saved_line

        # Display right sidebar
        self.new_column()
        for p in self._right_sidebar():
            if (hasattr(self.args, 'enable_' + p) or hasattr(self.args, 'disable_' + p)) and p in stat_display:
                self.new_line()
                if p in ['processlist', 'programlist']:
                    p_index = self._right_sidebar().index(p) + 1
                    self.display_plugin(
                        stat_display[p],
                        display_optional=(self.term_window.getmaxyx()[1] > 102),
                        display_additional=(not MACOS),
                        max_y=(
                            self.term_window.getmaxyx()[0]
                            - sum(
                                [
                                    self.get_stats_display_height(stat_display[i])
                                    for i in self._right_sidebar()[p_index:]
                                ]
                            )
                            - 2
                        ),
                    )
                else:
                    self.display_plugin(stat_display[p])

    def get_popup_size(self, message, size_x, size_y, popup_type, input_size):
        # Split the message into lines (seperated by \n)
        sentence_list = message.split('\n')
        # If length size_x has not been give, calculate the same
        # as a calculation of the max sentence length
        if size_x is None:
            size_x = len(max(sentence_list, key=len)) + 4
            # Add space for the input field
            if popup_type == 'input':
                size_x += input_size
        # if height size_y has not been given, calculate this depending on
        # the no of lines in sentence
        if size_y is None:
            size_y = len(sentence_list) + 4
        # Calculate screen size
        screen_x = self.term_window.getmaxyx()[1]
        screen_y = self.term_window.getmaxyx()[0]
        # Identify the centre position for the popup on the screen
        pos_x = int((screen_x - size_x) / 2)
        pos_y = int((screen_y - size_y) / 2)
        # return sentence_list message to be displayed, x & y pos of popup,
        # screen size in x, y and popup size in x, y
        return sentence_list, pos_x, pos_y, screen_x, screen_y, size_x, size_y

    def display_popup(
        self,
        message,
        size_x=None,
        size_y=None,
        duration=3,
        popup_type='info',
        input_size=30,
        input_value=None,
        is_password=False,
    ):
        """
        Display a centered popup.

        popup_type: ='info'
         Just an information popup, no user interaction
         Display a centered popup with the given message during duration seconds
         If size_x and size_y: set the popup size
         else set it automatically
         Return True if the popup could be displayed

        popup_type='input'
         Display a centered popup with the given message and a input field
         If size_x and size_y: set the popup size
         else set it automatically
         Return the input string or None if the field is empty

        popup_type='yesno'
         Display a centered popup with the given message
         If size_x and size_y: set the popup size
         else set it automatically
         Return True (yes) or False (no)
        """
        # Calculate the centre and content of the popup
        sentence_list, pos_x, pos_y, screen_x, screen_y, size_x, size_y = self.get_popup_size(
            message, size_x, size_y, popup_type, input_size
        )

        if size_x > screen_x or size_y > screen_y:
            # No size to display the popup => abort
            return False

        # Create the popup
        popup = curses.newwin(size_y, size_x, pos_y, pos_x)

        # Fill the popup
        popup.border()

        # Add the message
        for y, m in enumerate(sentence_list):
            if m:
                popup.addnstr(2 + y, 2, m, len(m))

        # Popup is for 'info' ie display only
        if popup_type == 'info':
            # Display the popup
            popup.refresh()
            self.wait(duration * 1000)
            return True

        # Popup accepts a input
        if popup_type == 'input':
            logger.info(popup_type)
            logger.info(is_password)
            return self._handle_input_popup(m, input_size, input_value, popup, is_password)

        # Popup accepts a user input as Y or N for yes no
        if popup_type == 'yesno':
            return self._handle_yesno_popup(popup, sentence_list, m)
        return None

    def _handle_input_popup(self, m, input_size, input_value, popup, is_password):
        # Create a single-line input sub-window positioned relative to the
        # rendered message length (m) within the parent popup
        sub_pop = popup.derwin(1, input_size, 2, 2 + len(m))
        sub_pop.attron(self.colors_list['FILTER'])

        # Pre-populate the field if an initial value is provided
        if input_value is not None:
            sub_pop.addnstr(0, 0, input_value, len(input_value))

        # Render both the popup and input field before capturing input
        popup.refresh()
        sub_pop.refresh()

        # Enable cursor visibility and keypad mode for user interaction
        self.set_cursor(2)
        self.term_window.keypad(1)

        # Password input: bypass curses textbox and use standard terminal input
        # (input is not displayed in the sub-window)
        if is_password:
            textbox = getpass.getpass('')
            self.set_cursor(0)
            if textbox != '':
                return textbox
            return None

        # Standard text input using curses textbox within the sub-window
        textbox = GlancesTextbox(sub_pop, insert_mode=True)
        textbox.edit()

        # Restore cursor state after input completes
        self.set_cursor(0)

        # Extract value; curses textbox appends a trailing newline, so strip it
        if textbox.gather() != '':
            return textbox.gather()[:-1]
        return None

    def _handle_yesno_popup(self, popup, sentence_list, m):
        # Create a 1x2 sub-window positioned after the last line of text;
        # width/offset is based on the rendered message length (m)
        sub_pop = popup.derwin(1, 2, len(sentence_list) + 1, len(m) + 2)

        # Apply input field styling (uses FILTER color scheme)
        sub_pop.attron(self.colors_list['FILTER'])

        # Initialize the field (defensive: curses may raise on small windows)
        try:
            sub_pop.addnstr(0, 0, '', 0)
        except curses.error:
            pass

        # Ensure both parent popup and input field are rendered
        popup.refresh()
        sub_pop.refresh()

        # Enable cursor + keypad for user input within the popup
        self.set_cursor(2)
        self.term_window.keypad(1)

        # Create and run the Yes/No textbox (single-line input)
        textbox = GlancesTextboxYesNo(sub_pop, insert_mode=False)
        textbox.edit()

        # Restore cursor state after input completes
        self.set_cursor(0)
        # Return the captured user input
        return textbox.gather()

    def setup_upper_left_pos(self, plugin_stats):
        screen_y, screen_x = self.term_window.getmaxyx()

        if plugin_stats['align'] == 'right':
            # Right align (last column)
            display_x = screen_x - self.get_stats_display_width(plugin_stats)
        else:
            display_x = self.column

        if plugin_stats['align'] == 'bottom':
            # Bottom (last line)
            display_y = screen_y - self.get_stats_display_height(plugin_stats)
        else:
            display_y = self.line

        return display_y, display_x

    def get_next_x_and_x_max(self, m, x, x_max):
        # New column
        # Python 2: we need to decode to get real screen size because
        # UTF-8 special tree chars occupy several bytes.
        # Python 3: strings are strings and bytes are bytes, all is
        # good.
        try:
            x += len(u(m['msg']))
        except UnicodeDecodeError:
            # Quick and dirty hack for issue #745
            pass
        if x > x_max:
            x_max = x

        return x, x_max

    def display_stats_with_current_size(self, m, y, x):
        screen_x = self.term_window.getmaxyx()[1]
        self.term_window.addnstr(
            y,
            x,
            m['msg'],
            # Do not display outside the screen
            screen_x - x,
            self.colors_list[m['decoration']],
        )

    def display_stats(self, plugin_stats, init, helper):
        y, x, x_max = init
        for m in plugin_stats['msgdict']:
            # New line
            try:
                if m['msg'].startswith('\n'):
                    y, x = helper['goto next, add first col'](y, x)
                    continue
            except Exception:
                # Avoid exception (see issue #1692)
                pass
            # Do not display outside the screen
            if x < 0:
                continue
            if helper['x overbound?'](m, x):
                continue
            if helper['y overbound?'](y):
                break
            # If display_optional = False do not display optional stats
            if helper['display optional?'](m):
                continue
            # If display_additional = False do not display additional stats
            if helper['display additional?'](m):
                continue
            # Is it possible to display the stat with the current screen size
            # !!! Crash if not try/except... Why ???
            try:
                self.display_stats_with_current_size(m, y, x)
            except Exception:
                pass
            else:
                x, x_max = self.get_next_x_and_x_max(m, x, x_max)

        return y, x, x_max

    def display_plugin(self, plugin_stats, display_optional=True, display_additional=True, max_y=65535, add_space=0):
        """Display the plugin_stats on the screen.

        :param plugin_stats:
        :param display_optional: display the optional stats if True
        :param display_additional: display additional stats if True
        :param max_y: do not display line > max_y
        :param add_space: add x space (line) after the plugin
        """
        # Exit if:
        # - the plugin_stats message is empty
        # - the display tag = False
        if plugin_stats is None or not plugin_stats['msgdict'] or not plugin_stats['display']:
            # Exit
            return 0

        # Get the screen size
        screen_y, screen_x = self.term_window.getmaxyx()

        # Set the upper/left position of the message
        display_y, display_x = self.setup_upper_left_pos(plugin_stats)

        helper = {
            'goto next, add first col': lambda y, x: (y + 1, display_x),
            'x overbound?': lambda m, x: not m['splittable'] and (x + len(m['msg']) > screen_x),
            'y overbound?': lambda y: y < 0 or (y + 1 > screen_y) or (y > max_y),
            'display optional?': lambda m: not display_optional and m['optional'],
            'display additional?': lambda m: not display_additional and m['additional'],
        }

        # Display
        init = display_y, display_x, display_x
        y, x, x_max = self.display_stats(plugin_stats, init, helper)

        # Compute the next Glances column/line position
        self.next_column = max(self.next_column, x_max + self.space_between_column)
        self.next_line = max(self.next_line, y + self.space_between_line)

        # Have empty lines after the plugins
        self.next_line += add_space
        return None

    def clear(self):
        """Erase the content of the screen.
        The difference is that clear() also calls clearok(). clearok()
        basically tells ncurses to forget whatever it knows about the current
        terminal contents, so that when refresh() is called, it will actually
        begin by clearing the entire terminal screen before redrawing any of it."""
        self.term_window.clear()

    def erase(self):
        """Erase the content of the screen.
        erase() on the other hand, just clears the screen (the internal
        object, not the terminal screen). When refresh() is later called,
        ncurses will still compute the minimum number of characters to send to
        update the terminal."""
        self.term_window.erase()

    def refresh(self):
        """Refresh the windows"""
        self.term_window.refresh()

    def flush(self, stats, cs_status=None):
        """Erase and update the screen.

        :param stats: Stats database to display
        :param cs_status:
            "None": standalone or server mode
            "Connected": Client is connected to the server
            "Disconnected": Client is disconnected from the server
        """
        # See https://stackoverflow.com/a/43486979/1919431
        self.erase()
        self.display(stats, cs_status=cs_status)
        self.refresh()

    def update(self, stats, duration=3, cs_status=None, return_to_browser=False):
        """Update the screen.

        :param stats: Stats database to display
        :param duration: duration of the loop
        :param cs_status:
            "None": standalone or server mode
            "Connected": Client is connected to the server
            "Disconnected": Client is disconnected from the server
        :param return_to_browser:
            True: Do not exist, return to the browser list
            False: Exit and return to the shell

        :return: True if exit key has been pressed else False
        """
        # Flush display
        self.flush(stats, cs_status=cs_status)

        # If the duration is < 0 (update + export time > refresh_time)
        # Then display the interface and log a message
        if duration <= 0:
            logger.warning('Update and export time higher than refresh_time.')
            duration = 0.1

        # Wait duration (in s) time
        isexitkey = False
        countdown = Timer(duration)
        # Set the default timeout (in ms) between two getch
        self.term_window.timeout(100)
        while not countdown.finished() and not isexitkey:
            # Getkey
            pressedkey = self.__catch_key(return_to_browser=return_to_browser)

            if pressedkey == -1:
                self.wait()
                continue

            isexitkey = pressedkey == ord('\x1b') or pressedkey == ord('q')

            if pressedkey == curses.KEY_F5 or self.pressedkey == 18:
                # Were asked to refresh (F5 or Ctrl-R)
                self.clear()
                return isexitkey

            if pressedkey in (curses.KEY_UP, 65, curses.KEY_DOWN, 66, curses.KEY_LEFT, 68, curses.KEY_RIGHT, 67):
                # Up of won key pressed, reset the countdown
                # Better for user experience
                countdown.reset()

            if isexitkey and self.args.help_tag:
                # Quit from help should return to main screen, not exit #1874
                self.args.help_tag = not self.args.help_tag
                return False

            if not isexitkey and pressedkey > -1:
                # Redraw display
                self.flush(stats, cs_status=cs_status)
                # Overwrite the timeout with the countdown
                self.wait(delay=int(countdown.get() * 1000))

        return isexitkey

    def wait(self, delay=100):
        """Wait delay in ms"""
        curses.napms(delay)

    def get_stats_display_width(self, curse_msg, without_option=False):
        """Return the width of the formatted curses message."""
        try:
            if without_option:
                # Size without options
                c = len(
                    max(
                        ''.join(
                            [
                                (u(u(nativestr(i['msg'])).encode('ascii', 'replace')) if not i['optional'] else "")
                                for i in curse_msg['msgdict']
                            ]
                        ).split('\n'),
                        key=len,
                    )
                )
            else:
                # Size with all options
                c = len(
                    max(
                        ''.join(
                            [u(u(nativestr(i['msg'])).encode('ascii', 'replace')) for i in curse_msg['msgdict']]
                        ).split('\n'),
                        key=len,
                    )
                )
        except Exception as e:
            logger.debug(f'ERROR: Can not compute plugin width ({e})')
            return 0
        else:
            return c

    def get_stats_display_height(self, curse_msg):
        """Return the height of the formatted curses message.

        The height is defined by the number of '\n' (new line).
        """
        try:
            c = [i['msg'] for i in curse_msg['msgdict']].count('\n')
        except Exception as e:
            logger.debug(f'ERROR: Can not compute plugin height ({e})')
            return 0
        else:
            return c + 1


class GlancesCursesStandalone(_GlancesCurses):
    """Class for the Glances curse standalone."""

    # Default number of processes to displayed is set to 50
    glances_processes.max_processes = 50


class GlancesCursesClient(_GlancesCurses):
    """Class for the Glances curse client."""

    # Default number of processes to displayed is set to 50
    # For the moment, cursor in client/server mode is not supported see #3221
    glances_processes.max_processes = 50


class GlancesTextbox(Textbox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def do_command(self, ch):
        if ch == 10:  # Enter
            return 0
        if ch == 127:  # Back
            return 8
        return super().do_command(ch)


class GlancesTextboxYesNo(Textbox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def do_command(self, ch):
        return super().do_command(ch)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

r"""
Default AMP
=========

Monitor a process by executing a command line. This is the default AMP's behavior
if no AMP script is found.

Configuration file example
--------------------------

[amp_foo]
enable=true
regex=\/usr\/bin\/foo
refresh=10
one_line=false
command=foo status
"""

from glances.amps.amp import GlancesAmp
from glances.logger import logger
from glances.secure import secure_popen


class Amp(GlancesAmp):
    """Glances' Default AMP."""

    NAME = ''
    VERSION = '1.1'
    DESCRIPTION = ''
    AUTHOR = 'Nicolargo'
    EMAIL = 'contact@nicolargo.com'

    def __init__(self, name=None, args=None):
        """Init the AMP."""
        self.NAME = name.capitalize()
        super().__init__(name=name, args=args)

    def update(self, process_list):
        """Update the AMP"""
        # Get the systemctl status
        logger.debug('{}: Update AMP stats using command {}'.format(self.NAME, self.get('service_cmd')))
        # Get command to execute
        try:
            res = self.get('command')
        except OSError as e:
            logger.debug(f'{self.NAME}: Error while executing command ({e})')
            return self.result()
        # No command found, use default message
        if res is None:
            # Set the default message if command return None
            # Default sum of CPU and MEM for the matching regex
            self.set_result(
                'CPU: {:.1f}% | MEM: {:.1f}%'.format(
                    sum([p['cpu_percent'] for p in process_list]), sum([p['memory_percent'] for p in process_list])
                )
            )
            return self.result()
        # Run command(s)
        # Comma separated commands can be executed
        try:
            self.set_result(secure_popen(res, allow_operators=self.allow_operators()).rstrip())
        except Exception as e:
            self.set_result(e.output)
        return self.result()



#!/usr/bin/env python
#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the Glances XML-RPC server security headers:
- Host header validation (GHSA-w856-8p3r-p338 / CVE-2026-46611)
- CORS per-request origin reflection (GHSA-87qc-fj39-wccr / CVE-2026-46608)."""

import shlex
import socket
import subprocess
import sys
import time
import unittest

import requests

SERVER_PORT = 62209
URL = f"http://127.0.0.1:{SERVER_PORT}/RPC2"

SECURE_PORT = 62210
SECURE_URL = f"http://127.0.0.1:{SECURE_PORT}/RPC2"

CORS_PORT = 62211
CORS_URL = f"http://127.0.0.1:{CORS_PORT}/RPC2"

XMLRPC_BODY = '<?xml version="1.0"?><methodCall><methodName>init</methodName></methodCall>'


pid = None
pid_secure = None
pid_cors = None


class TestGlancesXmlrpc(unittest.TestCase):
    """Glances XML-RPC server Host header validation tests."""

    def setUp(self):
        print('\n' + '=' * 78)

    def post(self, host_header):
        """POST an XML-RPC call with a specific Host header (default server)."""
        return requests.post(
            URL,
            data=XMLRPC_BODY,
            headers={'Host': host_header, 'Content-Type': 'text/plain'},
            timeout=5,
        )

    def post_secure(self, host_header):
        """POST an XML-RPC call with a specific Host header (allowlisted server)."""
        return requests.post(
            SECURE_URL,
            data=XMLRPC_BODY,
            headers={'Host': host_header, 'Content-Type': 'text/plain'},
            timeout=5,
        )

    def test_000_start_server(self):
        """Start the Glances XML-RPC server (no allowlist configured)."""
        global pid
        print('INFO: [TEST_000] Start the Glances XML-RPC Server')
        cmdline = sys.executable
        cmdline += f" -m glances -B 127.0.0.1 -s -p {SERVER_PORT} --disable-autodiscover -C ./conf/glances.conf"
        print(f"Run: {cmdline}")
        pid = subprocess.Popen(shlex.split(cmdline))
        print("Wait 5 seconds for server start...")
        time.sleep(5)
        self.assertIsNotNone(pid)

    def test_001_default_accepts_any_host(self):
        """Without xmlrpc_allowed_hosts set, any Host header is accepted."""
        print('INFO: [TEST_001] Default = permissive (no allowlist)')
        r = self.post('attacker.example.com')
        self.assertEqual(r.status_code, 200)
        self.assertIn('<methodResponse>', r.text)

    def test_010_start_secure_server(self):
        """Start a second XML-RPC server with xmlrpc_allowed_hosts configured."""
        global pid_secure
        print('INFO: [TEST_010] Start secured XML-RPC server')
        cmdline = sys.executable
        cmdline += (
            f" -m glances -B 127.0.0.1 -s -p {SECURE_PORT}"
            " --disable-autodiscover"
            " -C ./tests/conf/glances_xmlrpc_allowed_hosts.conf"
        )
        print(f"Run: {cmdline}")
        pid_secure = subprocess.Popen(shlex.split(cmdline))
        print("Wait 5 seconds for server start...")
        time.sleep(5)
        self.assertIsNotNone(pid_secure)

    def test_011_secure_rejects_spoofed_host(self):
        """With xmlrpc_allowed_hosts=127.0.0.1, Host: attacker.example.com -> 400."""
        print('INFO: [TEST_011] Spoofed Host rejected with 400')
        r = self.post_secure('attacker.example.com')
        self.assertEqual(r.status_code, 400)

    def test_012_secure_accepts_listed_host(self):
        """Allowlisted Host returns 200."""
        print('INFO: [TEST_012] Allowlisted Host accepted')
        r = self.post_secure('127.0.0.1')
        self.assertEqual(r.status_code, 200)
        self.assertIn('<methodResponse>', r.text)

    def test_013_secure_wildcard_match(self):
        """Wildcard pattern *.glances.test matches subdomain."""
        print('INFO: [TEST_013] Wildcard match')
        r = self.post_secure('node1.glances.test')
        self.assertEqual(r.status_code, 200)

    def test_014_secure_wildcard_no_bare_match(self):
        """*.glances.test does NOT match the bare domain glances.test."""
        print('INFO: [TEST_014] Wildcard does not match bare domain')
        r = self.post_secure('glances.test')
        self.assertEqual(r.status_code, 400)

    def test_015_secure_strips_port(self):
        """Host: 127.0.0.1:62210 matches the bare 127.0.0.1 entry."""
        print('INFO: [TEST_015] Port is stripped before matching')
        r = self.post_secure(f'127.0.0.1:{SECURE_PORT}')
        self.assertEqual(r.status_code, 200)

    def post_with_origin(self, url, origin):
        """POST an XML-RPC call with a specific Origin header."""
        return requests.post(
            url,
            data=XMLRPC_BODY,
            headers={'Origin': origin, 'Content-Type': 'text/plain'},
            timeout=5,
        )

    def test_020_default_cors_wildcard(self):
        """Default cors_origins=* echoes ACAO: * for any Origin."""
        print('INFO: [TEST_020] Default CORS wildcard')
        r = self.post_with_origin(URL, 'http://evil.example.com')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.headers.get('Access-Control-Allow-Origin'), '*')

    def test_030_start_cors_server(self):
        """Start a third XML-RPC server with a two-entry cors_origins allowlist."""
        global pid_cors
        print('INFO: [TEST_030] Start CORS-restricted XML-RPC server')
        cmdline = sys.executable
        cmdline += (
            f" -m glances -B 127.0.0.1 -s -p {CORS_PORT}"
            " --disable-autodiscover"
            " -C ./tests/conf/glances_xmlrpc_cors.conf"
        )
        print(f"Run: {cmdline}")
        pid_cors = subprocess.Popen(shlex.split(cmdline))
        print("Wait 5 seconds for server start...")
        time.sleep(5)
        self.assertIsNotNone(pid_cors)

    def test_031_cors_reflects_first_allowed_origin(self):
        """Multi-origin allowlist reflects a matching Origin, with Vary: Origin."""
        print('INFO: [TEST_031] First allowlisted Origin reflected')
        r = self.post_with_origin(CORS_URL, 'https://dashboard.glances.test')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(
            r.headers.get('Access-Control-Allow-Origin'),
            'https://dashboard.glances.test',
        )
        self.assertEqual(r.headers.get('Vary'), 'Origin')

    def test_032_cors_reflects_second_allowed_origin(self):
        """All entries in the allowlist are honoured, not just the first one."""
        print('INFO: [TEST_032] Second allowlisted Origin reflected')
        r = self.post_with_origin(CORS_URL, 'https://grafana.glances.test')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(
            r.headers.get('Access-Control-Allow-Origin'),
            'https://grafana.glances.test',
        )

    def test_033_cors_foreign_origin_no_header(self):
        """Foreign Origin: no ACAO header is emitted (request still succeeds)."""
        print('INFO: [TEST_033] Foreign Origin gets no ACAO header')
        r = self.post_with_origin(CORS_URL, 'http://evil.example.com')
        self.assertEqual(r.status_code, 200)
        self.assertNotIn('Access-Control-Allow-Origin', r.headers)

    def test_034_cors_no_origin_no_header(self):
        """Non-browser client (no Origin header): no ACAO emitted either."""
        print('INFO: [TEST_034] Missing Origin header -> no ACAO')
        r = requests.post(
            CORS_URL,
            data=XMLRPC_BODY,
            headers={'Content-Type': 'text/plain'},
            timeout=5,
        )
        self.assertEqual(r.status_code, 200)
        self.assertNotIn('Access-Control-Allow-Origin', r.headers)

    def test_016_secure_missing_host_rejected(self):
        """HTTP/1.0 request with no Host header -> 400."""
        print('INFO: [TEST_016] Missing Host header rejected')
        s = socket.create_connection(('127.0.0.1', SECURE_PORT), timeout=5)
        body = XMLRPC_BODY.encode()
        req = (
            b'POST /RPC2 HTTP/1.0\r\n'
            b'Content-Type: text/plain\r\n'
            b'Content-Length: ' + str(len(body)).encode() + b'\r\n\r\n' + body
        )
        s.sendall(req)
        resp = b''
        while True:
            chunk = s.recv(4096)
            if not chunk:
                break
            resp += chunk
        s.close()
        status_line = resp.split(b'\r\n', 1)[0]
        self.assertIn(b'400', status_line)

    def test_999_stop_server(self):
        """Stop all Glances XML-RPC servers."""
        print('INFO: [TEST_999] Stop all servers')
        pid.terminate()
        if pid_secure is not None:
            pid_secure.terminate()
        if pid_cors is not None:
            pid_cors.terminate()
        time.sleep(1)
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""JSON serialization utilities for Glances output layer."""

from datetime import datetime
from typing import Any

from glances.globals import json_dumps, json_loads
from glances.logger import logger


class PluginSerializationError:
    """Represents a serialization error for a plugin."""

    def __init__(self, plugin_name: str, error_message: str):
        self.plugin_name = plugin_name
        self.error_message = error_message

    def to_dict(self) -> dict[str, Any]:
        return {
            "error": True,
            "plugin": self.plugin_name,
            "message": self.error_message,
        }


class GlancesJSONSerializer:
    """Serializer that produces consistent, valid JSON output from plugin data."""

    def __init__(self, include_errors: bool = True, include_metadata: bool = False):
        """
        Initialize the JSON serializer.

        Args:
            include_errors: If True, include error information for failed plugins.
            include_metadata: If True, include metadata like timestamp.
        """
        self.include_errors = include_errors
        self.include_metadata = include_metadata

    def normalize_value(self, value: Any) -> Any:
        """Normalize a value to be JSON-serializable."""
        if value is None:
            return None

        if isinstance(value, bytes):
            return value.decode('utf-8', errors='replace')

        if isinstance(value, datetime):
            return value.isoformat()

        if isinstance(value, (int, float, str, bool)):
            return value

        if isinstance(value, dict):
            return {str(k): self.normalize_value(v) for k, v in value.items()}

        if isinstance(value, (list, tuple)):
            return [self.normalize_value(item) for item in value]

        if hasattr(value, '_asdict'):
            return self.normalize_value(value._asdict())

        return str(value)

    def serialize_plugin_data(self, plugin_name: str, raw_data: Any) -> dict[str, Any] | list[Any] | str | None:
        """
        Serialize plugin data to a JSON-compatible format.

        Args:
            plugin_name: Name of the plugin.
            raw_data: Raw data from the plugin (can be bytes, dict, list, etc.).

        Returns:
            JSON-compatible data structure, or error dict if serialization fails.
        """
        try:
            if raw_data is None:
                return None

            if isinstance(raw_data, bytes):
                decoded = raw_data.decode('utf-8', errors='replace')
                try:
                    return json_loads(decoded)
                except Exception:
                    return self.normalize_value(decoded)

            return self.normalize_value(raw_data)

        except Exception as e:
            logger.debug(f"Error serializing plugin {plugin_name}: {e}")
            if self.include_errors:
                return PluginSerializationError(plugin_name, str(e)).to_dict()
            return None

    def serialize_plugins(self, stats: Any, plugin_list: list[str] | None = None) -> dict[str, Any]:
        """
        Serialize data from multiple plugins into a single JSON object.

        Args:
            stats: The GlancesStats object.
            plugin_list: List of plugin names to include. If None, includes all enabled.

        Returns:
            Dictionary containing all plugin data.
        """
        result: dict[str, Any] = {}
        errors: list[dict[str, Any]] = []

        if plugin_list is None:
            plugin_list = stats.getPluginsList()

        for plugin_name in plugin_list:
            plugin_data = self._get_plugin_data(stats, plugin_name)
            if plugin_data is not None:
                if isinstance(plugin_data, dict) and plugin_data.get("error"):
                    errors.append(plugin_data)
                else:
                    result[plugin_name] = plugin_data

        if self.include_metadata:
            result["_metadata"] = {
                "timestamp": datetime.now().isoformat(),
                "plugin_count": len(result),
            }

        if self.include_errors and errors:
            result["_errors"] = errors

        return result

    def _get_plugin_data(self, stats: Any, plugin_name: str) -> Any:
        """Get and serialize data from a single plugin."""
        try:
            if plugin_name not in stats.getPluginsList():
                return None

            plugin = stats.get_plugin(plugin_name)
            if plugin is None or not plugin.is_enabled():
                return None

            json_data = plugin.get_json()
            return self.serialize_plugin_data(plugin_name, json_data)

        except Exception as e:
            logger.debug(f"Error getting data from plugin {plugin_name}: {e}")
            if self.include_errors:
                return PluginSerializationError(plugin_name, str(e)).to_dict()
            return None

    def to_json_string(self, data: Any) -> str:
        """Convert data to a JSON string."""
        try:
            result = json_dumps(data)
            if isinstance(result, bytes):
                return result.decode('utf-8')
            return result
        except Exception as e:
            logger.error(f"Error converting to JSON string: {e}")
            return '{"error": "Failed to serialize data"}'

    def serialize_to_string(self, stats: Any, plugin_list: list[str] | None = None) -> str:
        """
        Serialize plugins to a JSON string.

        Args:
            stats: The GlancesStats object.
            plugin_list: List of plugin names to include.

        Returns:
            JSON string containing all plugin data.
        """
        data = self.serialize_plugins(stats, plugin_list)
        return self.to_json_string(data)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Network plugin."""

import psutil

from glances.logger import logger
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: if True then compute and add *_gauge and *_rate_per_is fields
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'interface_name': {'description': 'Interface name.'},
    'alias': {'description': 'Interface alias name (optional).'},
    'bytes_recv': {
        'description': 'Number of bytes received.',
        'rate': True,
        'unit': 'byte',
    },
    'bytes_sent': {
        'description': 'Number of bytes sent.',
        'rate': True,
        'unit': 'byte',
    },
    'bytes_all': {
        'description': 'Number of bytes received and sent.',
        'rate': True,
        'unit': 'byte',
    },
    'speed': {
        'description': 'Maximum interface speed (in bit per second). Can return 0 on some operating-system.',
        'unit': 'bitpersecond',
    },
    'is_up': {'description': 'Is the interface up ?', 'unit': 'bool'},
}

# SNMP OID
# http://www.net-snmp.org/docs/mibs/interfaces.html
# Dict key = interface_name
snmp_oid = {
    'default': {
        'interface_name': '1.3.6.1.2.1.2.2.1.2',
        'bytes_recv': '1.3.6.1.2.1.2.2.1.10',
        'bytes_sent': '1.3.6.1.2.1.2.2.1.16',
    }
}

# Define the history items list
items_history_list = [
    {'name': 'bytes_recv_rate_per_sec', 'description': 'Download rate per second', 'y_unit': 'B/s'},
    {'name': 'bytes_sent_rate_per_sec', 'description': 'Upload rate per second', 'y_unit': 'B/s'},
]


class NetworkPlugin(GlancesPluginModel):
    """Glances network plugin.

    stats is a list
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(
            args=args,
            config=config,
            items_history_list=items_history_list,
            fields_description=fields_description,
            stats_init_value=[],
        )

        # We want to display the stat in the curse interface
        self.display_curse = True

        # Hide stats if it has never been != 0
        if config is not None:
            self.hide_zero = config.get_bool_value(self.plugin_name, 'hide_zero', default=False)
        else:
            self.hide_zero = False
        self.hide_zero_fields = ['bytes_recv_rate_per_sec', 'bytes_sent_rate_per_sec']

        #  Add support for automatically hiding network interfaces that are down
        # or that don't have any IP addresses #2799
        if config is not None:
            self.hide_no_up = config.get_bool_value(self.plugin_name, 'hide_no_up', default=False)
            self.hide_no_ip = config.get_bool_value(self.plugin_name, 'hide_no_ip', default=False)
        else:
            self.hide_no_up = False
            self.hide_no_ip = False

        # Force a first update because we need two updates to have the first stat
        self.update()
        self.refresh_timer.set(0)

    def get_key(self):
        """Return the key of the list."""
        return 'interface_name'

    # @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update network stats using the input method.

        :return: list of stats dict (one dict per interface)
        """
        if self.input_method == 'local':
            stats = self.update_local()

            # Update the stats
            for stat in stats:
                if stat.get('bytes_recv') is None:
                    print(f"def update: bytes_recv is None for {stat['interface_name']}")

        else:
            stats = self.get_init_value()

        self.stats = stats

        return self.stats

    @GlancesPluginModel._manage_rate
    def update_local(self):
        # Update stats using the standard system lib

        stats = self.get_init_value()

        # Grab network interface stat using the psutil net_io_counter method
        # Example:
        # { 'veth4cbf8f0a': snetio(
        #   bytes_sent=102038421, bytes_recv=1263258,
        #   packets_sent=25046, packets_recv=14114,
        #   errin=0, errout=0, dropin=0, dropout=0), ... }
        try:
            net_io_counters = psutil.net_io_counters(pernic=True)
        except UnicodeDecodeError as e:
            logger.debug(f'Can not get network interface counters ({e})')
            return self.stats

        # Grab interface's status (issue #765)
        # Grab interface's speed (issue #718)
        # Example:
        # { 'veth4cbf8f0a': snicstats(
        #   isup=True, duplex=<NicDuplex.NIC_DUPLEX_FULL: 2>,
        #   speed=10000, mtu=1500, flags='up,broadcast,running,multicast'), ... }
        net_status = {}
        try:
            net_status = psutil.net_if_stats()
            net_addrs = psutil.net_if_addrs()
        except OSError as e:
            # see psutil #797/glances #1106
            logger.debug(f'Can not get network interface status ({e})')

        # Filter interfaces (related to #2799)
        if self.hide_no_up:
            net_status = {k: v for k, v in net_status.items() if v.isup}
        if self.hide_no_ip:
            net_status = {
                k: v
                for k, v in net_status.items()
                if k in net_addrs and any(a.family != psutil.AF_LINK for a in net_addrs[k])
            }

        for interface_name, interface_stat in net_io_counters.items():
            # Do not take hidden interface into account
            # or KeyError: 'eth0' when interface is not connected #1348
            if not self.is_display(interface_name) or interface_name not in net_status:
                continue

            # Filter stats to keep only the fields we want (define in fields_description)
            # It will also convert psutil objects to a standard Python dict
            stat = self.filter_stats(interface_stat)
            stat.update(self.filter_stats(net_status[interface_name]))

            # Add the key
            stat['key'] = self.get_key()

            # Add disk name
            stat['interface_name'] = interface_name

            # Add alias define in the configuration file
            stat['alias'] = self.has_alias(interface_name)

            # Add sent + revc  stats
            stat['bytes_all'] = stat['bytes_sent'] + stat['bytes_recv']

            # Interface speed in Mbps, convert it to bps
            # Can be always 0 on some OSes
            stat['speed'] = stat['speed'] * 1048576

            stats.append(stat)

        return stats

    def update_views(self):
        """Update stats views."""
        # Call the father's method
        super().update_views()

        # Add specifics information
        # Alert
        for i in self.get_raw():
            if_real_name = i['interface_name'].split(':')[0]

            # Skip alert if no timespan to measure
            if not i.get('bytes_recv_rate_per_sec') or not i.get('bytes_sent_rate_per_sec'):
                continue

            # Convert rate to bps (to be able to compare to interface speed)
            bps_rx = int(i['bytes_recv_rate_per_sec'] * 8)
            bps_tx = int(i['bytes_sent_rate_per_sec'] * 8)

            # Decorate the bitrate with the configuration file thresholds
            alert_rx = self.get_alert(bps_rx, header='rx', action_key=if_real_name)
            alert_tx = self.get_alert(bps_tx, header='tx', action_key=if_real_name)

            # If nothing is define in the configuration file...
            # ... then use the interface speed (not available on all systems)
            if alert_rx == 'DEFAULT' and 'speed' in i and i['speed'] != 0:
                alert_rx = self.get_alert(current=bps_rx, maximum=i['speed'], header='rx')
            if alert_tx == 'DEFAULT' and 'speed' in i and i['speed'] != 0:
                alert_tx = self.get_alert(current=bps_tx, maximum=i['speed'], header='tx')

            # then decorates
            self.views[i[self.get_key()]]['bytes_recv']['decoration'] = alert_rx
            self.views[i[self.get_key()]]['bytes_recv_rate_per_sec']['decoration'] = alert_rx
            self.views[i[self.get_key()]]['bytes_sent']['decoration'] = alert_tx
            self.views[i[self.get_key()]]['bytes_sent_rate_per_sec']['decoration'] = alert_rx

    def _msg_curse_header(self, args, name_max_width):
        """Return the header curse lines."""
        ret = [self.curse_add_line('{:{width}}'.format('NETWORK', width=name_max_width), "TITLE")]
        if args and args.network_cumul:
            if args and args.network_sum:
                ret.append(self.curse_add_line('{:>14}'.format('Rx+Tx')))
            else:
                ret.append(self.curse_add_line('{:>7}'.format('Rx')))
                ret.append(self.curse_add_line('{:>7}'.format('Tx')))
        else:
            if args and args.network_sum:
                ret.append(self.curse_add_line('{:>14}'.format('Rx+Tx/s')))
            else:
                ret.append(self.curse_add_line('{:>7}'.format('Rx/s')))
                ret.append(self.curse_add_line('{:>7}'.format('Tx/s')))
        return ret

    def _get_if_name(self, i, name_max_width):
        """Return the (possibly truncated) display name for an interface."""
        if_name = i['alias'] if i['alias'] is not None else i['interface_name'].split(':')[0]
        if len(if_name) > name_max_width:
            # Cut interface name if it is too long
            if_name = '_' + if_name[-name_max_width + 1 :]
        return if_name

    def _get_if_rates(self, i, args):
        """Return (rx, tx, ax) formatted strings, or None if rates are unavailable.

        Returns None when a new interface appears before its first rate measurement.
        """
        to_bit, unit = (1, '') if (args and args.byte) else (8, 'b')
        if args and args.network_cumul:
            if i.get('bytes_recv') is None or i.get('bytes_sent') is None:
                return None
            rx = self.auto_unit(int(i['bytes_recv'] * to_bit)) + unit
            tx = self.auto_unit(int(i['bytes_sent'] * to_bit)) + unit
            ax = self.auto_unit(int(i['bytes_all'] * to_bit)) + unit
        else:
            if i.get('bytes_recv_rate_per_sec') is None or i.get('bytes_sent_rate_per_sec') is None:
                return None
            rx = self.auto_unit(int(i['bytes_recv_rate_per_sec'] * to_bit)) + unit
            tx = self.auto_unit(int(i['bytes_sent_rate_per_sec'] * to_bit)) + unit
            ax = self.auto_unit(int(i['bytes_all_rate_per_sec'] * to_bit)) + unit
        return rx, tx, ax

    def _msg_curse_if_line(self, i, if_name, rx, tx, ax, args, name_max_width):
        """Return curse lines for a single interface row."""
        ret = [
            self.curse_new_line(),
            self.curse_add_line('{:{width}}'.format(if_name, width=name_max_width)),
        ]
        if args and args.network_sum:
            ret.append(self.curse_add_line(f'{ax:>14}'))
        else:
            ret.append(
                self.curse_add_line(
                    f'{rx:>7}', self.get_views(item=i[self.get_key()], key='bytes_recv', option='decoration')
                )
            )
            ret.append(
                self.curse_add_line(
                    f'{tx:>7}', self.get_views(item=i[self.get_key()], key='bytes_sent', option='decoration')
                )
            )
        return ret

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        ret = []

        if not self.stats or self.is_disabled():
            return ret

        if not max_width:
            logger.debug(f"No max_width defined for the {self.plugin_name} plugin, it will not be displayed.")
            return ret

        name_max_width = max_width - 12
        ret.extend(self._msg_curse_header(args, name_max_width))

        for i in self.sorted_stats():
            # Do not display interface in down state (issue #765)
            if ('is_up' in i) and (i['is_up'] is False):
                continue
            # Hide stats if never be different from 0 (issue #1787)
            if all(self.get_views(item=i[self.get_key()], key=f, option='hidden') for f in self.hide_zero_fields):
                continue

            rates = self._get_if_rates(i, args)
            if rates is None:
                # Avoid issue when a new interface is created on the fly
                # Example: start Glances, then start a new container
                continue

            if_name = self._get_if_name(i, name_max_width)
            ret.extend(self._msg_curse_if_line(i, if_name, *rates, args, name_max_width))

        return ret



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""TimescaleDB interface class."""

import sys
import time
from datetime import datetime, timezone
from platform import node

import psycopg
from psycopg import sql

from glances.exports.export import GlancesExport
from glances.logger import logger

# Define the type conversions for TimescaleDB
# https://www.postgresql.org/docs/current/datatype.html
convert_types = {
    'bool': 'BOOLEAN',
    'int': 'BIGINT',
    'float': 'DOUBLE PRECISION',
    'str': 'TEXT',
    'tuple': 'TEXT',  # Store tuples as TEXT (comma-separated)
    'list': 'TEXT',  # Store lists as TEXT (comma-separated)
    'NoneType': 'DOUBLE PRECISION',  # Use DOUBLE PRECISION for NoneType to avoid issues with NULL
}


class Export(GlancesExport):
    """This class manages the TimescaleDB export module."""

    def __init__(self, config=None, args=None):
        """Init the TimescaleDB export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        self.db = None

        # Optional configuration keys
        self.user = None
        self.password = None
        self.hostname = None

        # Load the configuration file
        self.export_enable = self.load_conf(
            'timescaledb', mandatories=['host', 'port', 'db'], options=['user', 'password', 'hostname']
        )
        if not self.export_enable:
            exit('Missing TimescaleDB config')

        # The hostname is always add as an identifier in the TimescaleDB table
        # so we can filter the stats by hostname
        self.hostname = self.hostname or node().split(".")[0]

        # Init the TimescaleDB client
        self.client = self.init()

    def init(self):
        """Init the connection to the TimescaleDB server."""
        if not self.export_enable:
            return None

        try:
            # See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING
            conn_str = f"host={self.host} port={self.port} dbname={self.db} user={self.user} password={self.password}"
            db = psycopg.connect(conn_str)
        except Exception as e:
            logger.critical(f"Cannot connect to TimescaleDB server {self.host}:{self.port} ({e})")
            sys.exit(2)
        else:
            logger.info(f"Stats will be exported to TimescaleDB server: {self.host}:{self.port}")

        return db

    def normalize(self, value):
        """Normalize the value for use in a parameterized psycopg query (returns raw Python value)."""
        if isinstance(value, (list, tuple)):
            # Special case for list of one boolean
            if len(value) == 1 and isinstance(value[0], bool):
                return value[0]
            return ', '.join(str(v) for v in value)
        return value  # None → NULL, bool/str/int/float handled natively by psycopg

    def update(self, stats):
        """Update the TimescaleDB export module."""
        if not self.export_enable:
            return False

        # Get all the stats & limits
        # @TODO: Current limitation with sensors, fs and diskio plugins because fields list is not the same
        self._last_exported_list = [p for p in self.plugins_to_export(stats) if p not in ['sensors', 'fs', 'diskio']]
        all_stats = stats.getAllExportsAsDict(plugin_list=self.last_exported_list())
        all_limits = stats.getAllLimitsAsDict(plugin_list=self.last_exported_list())

        # Loop over plugins to export
        for plugin in self.last_exported_list():
            if isinstance(all_stats[plugin], dict):
                all_stats[plugin].update(all_limits[plugin])
                # Remove the <plugin>_disable field
                all_stats[plugin].pop(f"{plugin}_disable", None)
                # user is a special field that should not be exported
                # rename it to user_<plugin>
                if 'user' in all_stats[plugin]:
                    all_stats[plugin][f'user_{plugin}'] = all_stats[plugin].pop('user')
            elif isinstance(all_stats[plugin], list):
                for i in all_stats[plugin]:
                    i.update(all_limits[plugin])
                    # Remove the <plugin>_disable field
                    i.pop(f"{plugin}_disable", None)
                    # user is a special field that should not be exported
                    # rename it to user_<plugin>
                    if 'user' in i:
                        i[f'user_{plugin}'] = i.pop('user')
            else:
                continue

            plugin_stats = all_stats[plugin]
            creation_list = []  # List used to create the TimescaleDB table
            segmented_by = []  # List of columns used to segment the data
            values_list = []  # List of values to insert (list of lists, one list per row)
            if isinstance(plugin_stats, dict):
                # Stats is a dict
                # Create the list used to create the TimescaleDB table
                creation_list.append('time TIMESTAMPTZ NOT NULL')
                creation_list.append('hostname_id TEXT NOT NULL')
                segmented_by.extend(['hostname_id'])  # Segment by hostname
                for key, value in plugin_stats.items():
                    creation_list.append(f"{key} {convert_types[type(value).__name__]} NULL")
                values_list.append(datetime.now(timezone.utc))  # Add the current time (insertion time)
                values_list.append(self.hostname)  # Add the hostname
                values_list.extend([self.normalize(value) for value in plugin_stats.values()])
                values_list = [values_list]
            elif isinstance(plugin_stats, list) and len(plugin_stats) > 0 and 'key' in plugin_stats[0]:
                # Stats is a list
                # Create the list used to create the TimescaleDB table
                creation_list.append('time TIMESTAMPTZ NOT NULL')
                creation_list.append('hostname_id TEXT NOT NULL')
                creation_list.append('key_id TEXT NOT NULL')
                segmented_by.extend(['hostname_id', 'key_id'])  # Segment by hostname and key
                for key, value in plugin_stats[0].items():
                    # The 'key' field is already exported as key_id, do not duplicate it
                    if key == 'key':
                        continue
                    creation_list.append(f"{key} {convert_types[type(value).__name__]} NULL")
                # Create the values list (it is a list of list to have a single datamodel for all the plugins)
                for plugin_item in plugin_stats:
                    item_list = []
                    item_list.append(datetime.now(timezone.utc))  # Add the current time (insertion time)
                    item_list.append(self.hostname)  # Add the hostname
                    item_list.append(plugin_item.get('key'))
                    item_list.extend([self.normalize(value) for key, value in plugin_item.items() if key != 'key'])
                    values_list.append(item_list)
            else:
                continue

            # Export stats to TimescaleDB
            # logger.info(plugin)
            # logger.info(f"Segmented by: {segmented_by}")
            # logger.info(list(zip(creation_list, values_list[0])))
            self.export(plugin, creation_list, segmented_by, values_list)

        return True

    def export(self, plugin, creation_list, segmented_by, values_list):
        """Export the stats to the TimescaleDB server."""
        logger.debug(f"Export {plugin} stats to TimescaleDB")

        with self.client.cursor() as cur:
            # Is the table exists?
            cur.execute(
                "SELECT EXISTS(SELECT * FROM information_schema.tables WHERE table_name=%s)",
                [plugin],
            )
            if not cur.fetchone()[0]:
                # Create the table if it does not exist
                # https://github.com/timescale/timescaledb/blob/main/README.md#create-a-hypertable
                # Build CREATE TABLE using sql.Identifier for column names (prevents injection)
                # Each item in creation_list is "colname TYPE [NULL|NOT NULL]"
                fields = sql.SQL(', ').join(
                    sql.SQL("{} {}").format(sql.Identifier(item.split(' ')[0]), sql.SQL(' '.join(item.split(' ')[1:])))
                    for item in creation_list
                )
                create_query = sql.SQL(
                    "CREATE TABLE {table} ({fields}) WITH ("
                    "timescaledb.hypertable, "
                    "timescaledb.partition_column='time', "
                    "timescaledb.segmentby = {segmentby});"
                ).format(
                    table=sql.Identifier(plugin),
                    fields=fields,
                    segmentby=sql.Literal(', '.join(segmented_by)),
                )
                logger.debug(f"Create table: {create_query}")
                try:
                    cur.execute(create_query)
                except Exception as e:
                    logger.error(f"Cannot create table {plugin}: {e}")
                    return

            # Insert the data using parameterized queries (prevents injection)
            # https://github.com/timescale/timescaledb/blob/main/README.md#insert-and-query-data
            col_names = [item.split(' ')[0] for item in creation_list]
            cols = sql.SQL(', ').join(sql.Identifier(c) for c in col_names)
            placeholders = sql.SQL(', ').join(sql.Placeholder() for _ in col_names)
            insert_query = sql.SQL("INSERT INTO {table} ({cols}) VALUES ({vals})").format(
                table=sql.Identifier(plugin),
                cols=cols,
                vals=placeholders,
            )
            logger.debug(f"Insert data into table: {insert_query}")
            try:
                cur.executemany(insert_query, values_list)
            except Exception as e:
                logger.error(f"Cannot insert data into table {plugin}: {e}")
                return

        # Commit the changes (for every plugin or to be done at the end ?)
        self.client.commit()

    def exit(self):
        """Close the TimescaleDB export module."""
        # Force last write
        self.client.commit()

        # Close the TimescaleDB client
        time.sleep(3)  # Wait a bit to ensure all data is written
        self.client.close()

        # Call the father method
        super().exit()



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2026 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#


from dataclasses import dataclass


@dataclass
class NPUStats:
    """Container for NPU statistics"""

    def __init__(self):
        self.npu_id: str = "unknown"
        self.name: str = "unknown"
        self.load: int | None = None
        self.freq: int | None = None
        self.freq_current: int | None = None
        self.freq_max: int | None = None
        self.mem: int | None = None
        self.memory_used: int | None = None
        self.memory_total: int | None = None
        self.temperature: float | None = None
        self.power: float | None = None



####################################################################################
#
# Migration of the code issue3290.py to use multiprocessing with 'fork' start method
#
####################################################################################

import multiprocessing
import time

import psutil

# multiprocessing.set_start_method("fork")
ctx_mp_fork = multiprocessing.get_context('fork')


def exit_after(seconds, default=None):
    """Exit the function if it takes more than 'second' seconds to complete.
    In this case, return the value of 'default' (default: None)."""

    def handler(q, func, args, kwargs):
        q.put(func(*args, **kwargs))

    def decorator(func):
        def wraps(*args, **kwargs):
            q = ctx_mp_fork.Queue()
            p = ctx_mp_fork.Process(target=handler, args=(q, func, args, kwargs))
            p.start()
            p.join(timeout=seconds)
            if not p.is_alive():
                return q.get()

            p.terminate()
            p.join(timeout=0.1)
            if p.is_alive():
                # Kill in case processes doesn't terminate
                # Happens with cases like broken NFS connections
                p.kill()
            return default

        return wraps

    return decorator


class Issue3290:
    @exit_after(1, default=None)
    def blocking_io_call(self, fs):
        try:
            return psutil.disk_usage(fs)
        except OSError:
            return None


issue = Issue3290()
while True:
    print(f"{time.time()} {issue.blocking_io_call('/home/nicolargo/tmp/hang')}")
    time.sleep(1)
